#include <string>
#include <list>
#include <fstream>
#include <algorithm>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

#include "File.hpp"

using namespace std;

lockdown::io::File::File(const std::string& file, bool createIfNotFound) : filename(file) {
	if(stat(filename.c_str(), &fileInfo) < 0)  {
		if(errno != ENOENT)		//It does exist
			throw OpenError(filename);
		if(!createIfNotFound)	//Should I create it?
			throw NotFoundError(filename);
		fstream f(filename.c_str(), ios::out | ios::trunc);
		f.close();
		if(stat(filename.c_str(), &fileInfo) < 0) 
			throw OpenError(filename);
	}
}

void lockdown::io::File::read(std::list<std::string>& readTo) {
	ifstream o;
	o.open(filename.c_str());
	if(!o) 
		throw OpenError(filename);

	string input;
	while(o.good()) {
		getline(o, input);
		readTo.push_back(input);
	}
	if(!o.eof()) {
		o.close();	
		throw ReadError(filename);
	}
	o.close();	
}

void lockdown::io::File::write(const std::list<std::string>& writeFrom) {
	ofstream f;
	f.open(filename.c_str(), ios::trunc);
	if(!f) 
		throw OpenError(filename);

	list<string>::const_iterator index = writeFrom.begin();
	while(index != writeFrom.end()){
		f<<*index;
		index++;
		if(index != writeFrom.end())
			f<<'\n';
	}
	if(!f) {
		f.close();	
		throw WriteError(filename);
	}
	f.close();
}

void lockdown::io::File::setPermissions(std::string& perms) {
	bool minus = false;
	if(perms.at(0) == '-') {		//Setting the absolute permissions
		minus = true;
		perms.erase(0, 1);		//Remove the '-'
	}
	
	reverse(perms.begin(), perms.end());
	
	mode_t d = 0;
	for(int i = perms.size()-1; i >= 0;i--) {
		switch(perms.at(i)) {
			case '0':
				continue;
			case '7':
				if(i == 0)
					d += S_IRWXO;
				else if(i == 1)
					d += S_IRWXG;
				else if(i == 2)
					d += S_IRWXU;
				else if(i == 3) {
					d += S_ISUID;
					d += S_ISGID;
					d += S_ISVTX;
				}
				break;
			case '6':
				if(i == 0) {
					d += S_IROTH;
					d += S_IWOTH;
				}
				else if(i == 1) {
					d += S_IRGRP;
					d += S_IWGRP;
				}
				else if(i == 2) {
					d += S_IRGRP;
					d += S_IWGRP;
				}
				else if(i == 3) {
					d += S_ISUID;
					d += S_ISGID;
				}
				break;
			case '5':
				if(i == 0) {
					d += S_IROTH;
					d += S_IXOTH;
				}
				else if(i == 1) {
					d += S_IRGRP;
					d += S_IXGRP;
				}
				else if(i == 2) {
					d += S_IRGRP;
					d += S_IXGRP;
				}
				else if(i == 3) {
					d += S_ISUID;
					d += S_ISVTX;
				}
				break;
			case '4':
				if(i == 0)
					d += S_IROTH;
				else if(i == 1)
					d += S_IRGRP;
				else if(i == 2)
					d += S_IRGRP;
				else if(i == 3) 
					d += S_ISUID;
				break;
			case '3':	//falling
			case '2':
				if(i == 0)
					d += S_IWOTH;
				else if(i == 1)
					d += S_IWGRP;
				else if(i == 2)
					d += S_IWGRP;
				else if(i == 3) 
					d += S_ISGID;
				break;
			case '1':
				if(i == 0)
					d += S_IXOTH;
				else if(i == 1)
					d += S_IXGRP;
				else if(i == 2)
					d += S_IXGRP;
				else if(i == 3)
					d += S_ISVTX;
				break;
			default:
				string error = filename + ": unknown permission: " + perms;
				throw PermissionError(error);
		}
	}
	
	if(minus) {			//Remove the bits
		mode_t p = fileInfo.st_mode;	//Current mode
		//If d has the following, remove them from p
		if(d & S_IRUSR && p & S_IRUSR)		// R for owner
			p -= S_IRUSR;
		if(d & S_IWUSR && p & S_IWUSR)		// W for owner
			p -= S_IWUSR;
		if(d & S_IXUSR && p & S_IXUSR)		// X for owner
			p -= S_IXUSR;
		if(d & S_IRGRP && p & S_IRGRP)		// R for group
			p -= S_IRGRP;
		if(d & S_IWGRP && p & S_IWGRP)		// W for group
			p -= S_IWGRP;
		if(d & S_IXGRP && p & S_IXGRP)		// X for group
			p -= S_IXGRP;
		if(d & S_IROTH && p & S_IROTH)		// R for other
			p -= S_IROTH;
		if(d & S_IWOTH && p & S_IWOTH)		// W for other
			p -= S_IWOTH;
		if(d & S_IXOTH && p & S_IXOTH)		// X for other
			p -= S_IXOTH;
		if(d & S_ISUID && p & S_ISUID)		// user id on execution
			p -= S_ISUID;
		if(d & S_ISGID && p & S_ISGID)		// group id on execution
			p -= S_ISGID;
		if(d & S_ISVTX && p & S_ISVTX)		// sticky bit
			p -= S_ISVTX;

		d = p;
	}
	
	if(chmod(filename.c_str(), d) < 0) 
		throw PermissionError(filename + ": Error setting permissions: " + perms);
}

void lockdown::io::File::setUser(const std::string& user) {
	uid_t userid = fileInfo.st_uid;
	gid_t groupid = fileInfo.st_gid;
	
	passwd *info = getpwnam(user.c_str());
	if(info == 0) {
		string error = filename + ": unknown user: " + user;
		throw NotFoundError(filename);
	}
	userid = info->pw_uid;
	
	if(chown(filename.c_str(), userid, groupid) < 0) {
		string error = filename + ": Error setting user: " + user;
		throw PermissionError(error);
	}
}

void lockdown::io::File::setGroup(const std::string& groupname) {
	uid_t userid = fileInfo.st_uid;
	gid_t groupid = fileInfo.st_gid;
	
	group *gr = getgrnam(groupname.c_str());
	if(gr == 0) {
		string error = filename + ": unknown group: " + groupname;
		throw NotFoundError(filename);
	}
	groupid = gr->gr_gid;

	if(chown(filename.c_str(), userid, groupid) < 0) {
		string error = filename + ": Error setting group: " + groupname;
		throw PermissionError(error);
	}
}

void lockdown::io::File::setFlags(const std::string& flags) {
	list<string> theFlags;
	
	string::size_type pos1 = 0;
	string::size_type pos2 = 0;
	while((pos2 = flags.find(',', pos1)) != string::npos) {	//Save all the flags in "theFlags"
		theFlags.push_back(flags.substr(pos1, pos2 - pos1));
		pos1 = pos2 + 1;
	}
	theFlags.push_back(flags.substr(pos1, flags.size() - pos1));		//Don't forget the last (or only) flag
	
	unsigned long f = 0;
	
	list<string>::iterator finder = theFlags.begin();
	for(;finder != theFlags.end();finder++) {
		if(*finder == "nodump") {
			f |= UF_NODUMP;
		}
		else if(*finder == "uchg" || *finder == "uchange"  || *finder == "uimmutable") {
			f |= UF_IMMUTABLE;
		}
		else if(*finder == "uappnd" || *finder == "uappend") {
			f |= UF_APPEND;
		}
		else if(*finder == "uunlnk" || *finder =="uunlink") {
			f |= UF_NOUNLINK;
		}
		else if(*finder == "opaque") {
			f |= UF_OPAQUE;
		}
		else if(*finder == "arch" || *finder== "archived") {
			f |= SF_ARCHIVED;
		}
		else if(*finder == "schg" || *finder =="schange" || *finder =="simmutable") {
			f |= SF_IMMUTABLE;
		}
		else if(*finder == "sappnd" || *finder =="sappend") {
			f |= SF_APPEND;
		}
		else if(*finder == "sunlnk"|| *finder =="sunlink") {
			f |= SF_NOUNLINK;
		}
		else {
			string error = filename + ": unknown flag " + *finder;
			throw NotFoundError(error);
		}
	}
	
	if(chflags(filename.c_str(), f) < 0) {
		string error = filename + ": error setting flags: " + flags;
		throw PermissionError(error);
	}
}
